use std::{cmp::Reverse, collections::HashMap, usize};

use log::warn;

use crate::solutions::day7::second_ordering;

use super::{dealt_hand::DealtHand, hand_kind::HandKind};

#[derive(Debug, PartialEq, Eq)]
pub struct CategorizedHand {
    kind: HandKind,
    rest: DealtHand,
}

impl CategorizedHand {
    pub fn rest(&self) -> &DealtHand {
        &self.rest
    }
    pub fn bid(&self) -> usize {
        self.rest().bid()
    }
}

impl Ord for CategorizedHand {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        match self.kind.cmp(&other.kind) {
            std::cmp::Ordering::Equal => {
                assert!(
                    self.kind == other.kind,
                    "If ordering is equal then the hand kind must be the same too\n\
                    Self kind: ({:?}) and other kind: ({:?})",
                    self.kind,
                    other.kind
                );
                second_ordering::compare_along_str(self.rest().hand(), other.rest().hand())
            }
            not_equal_ordering => {
                assert!(
                    self.kind != other.kind,
                    "If ordering is not equal then the hand kind must not be the same either\n\
                    Self kind: ({:?}) and other kind: ({:?})",
                    self.kind,
                    other.kind
                );
                not_equal_ordering
            }
        }
    }
}

impl PartialOrd for CategorizedHand {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

impl From<DealtHand> for CategorizedHand {
    fn from(value: DealtHand) -> Self {
        fn calc_kind(values: &[usize]) -> Option<HandKind> {
            let max = values.first()?;

            let kind = match max {
                5 => HandKind::Five,
                4 => HandKind::Four,
                under_five_or_four => {
                    let second_after_max = values.get(1)?;
                    match (under_five_or_four, second_after_max) {
                        (3, 2) => HandKind::FullHouse,
                        (3, _) => HandKind::Three,
                        (2, 2) => HandKind::TwoPair,
                        (2, 1) => HandKind::OnePair,
                        _no_duplicates => HandKind::HighCard,
                    }
                }
            };

            Some(kind)
        }

        fn calc_duplicate_counted(value: &DealtHand) -> Vec<usize> {
            const FIRST_ENCOUNTERED: usize = 1;

            let mut count_of_same: HashMap<char, usize> = HashMap::with_capacity(10);

            for next_char in value.hand().chars() {
                count_of_same
                    .entry(next_char)
                    .and_modify(|count| *count = *count + 1)
                    .or_insert(FIRST_ENCOUNTERED);
            }

            {
                let mut to_sort: Vec<usize> =
                    count_of_same.into_iter().map(|(_, count)| count).collect();
                to_sort.sort_by_key(|&count| Reverse(count));
                to_sort
            }
        }

        let duplicated_counted = calc_duplicate_counted(&value);

        let kind = calc_kind(&duplicated_counted).unwrap_or_else(|| {
            warn!(
                "Got hand ({:?}) with less than 2 cards",
                &duplicated_counted
            );
            let default_value = HandKind::default();
            warn!("Default kind ({:?}) is assumed", default_value);
            default_value
        });

        CategorizedHand { kind, rest: value }
    }
}

#[cfg(test)]
mod testing {
    use std::cmp::Ordering;

    use crate::solutions::day7::{
        categorized_hand::CategorizedHand, dealt_hand::DealtHand, hand_kind::HandKind,
    };

    #[test]
    fn from_dealt_hand_to_categorized_hand() {
        fn assert_case(given: DealtHand, expected: HandKind) {
            let expected_hand = CategorizedHand {
                kind: expected,
                rest: given.clone(),
            };
            let actual = given.into();
            assert_eq!(expected_hand, actual);
        }

        assert_case(DealtHand::new("AAAAA", 20), HandKind::Five);
        assert_case(DealtHand::new("AA8AA", 20), HandKind::Four);
        assert_case(DealtHand::new("23332", 20), HandKind::FullHouse);
        assert_case(DealtHand::new("TTT98", 20), HandKind::Three);
        assert_case(DealtHand::new("23456", 20), HandKind::HighCard);
        assert_case(DealtHand::new("23432", 20), HandKind::TwoPair);
        assert_case(DealtHand::new("A23A4", 20), HandKind::OnePair);
    }

    #[test]
    fn compare_correctly_by_kind_and_second_ordering() {
        fn assert_case(left: DealtHand, right: DealtHand, expected: Ordering) {
            let (left, right): (CategorizedHand, CategorizedHand) = (left.into(), right.into());
            let actual = left.cmp(&right);
            assert_eq!(expected, actual, "Left: {:?}\nRight: {:?}", left, right);
        }

        assert_case(
            DealtHand::new("AAAAA", 20),
            DealtHand::new("BBBBA", 30),
            Ordering::Greater,
        );
        assert_case(
            DealtHand::new("12345", 20),
            DealtHand::new("BBBBA", 30),
            Ordering::Less,
        );
        assert_case(
            DealtHand::new("12345", 20),
            DealtHand::new("28495", 30),
            Ordering::Less,
        );
        assert_case(
            DealtHand::new("QQQJA", 20),
            DealtHand::new("T55J5", 30),
            Ordering::Greater,
        );
        //KK677 and KTJJT are both two pair.
        //Their first cards both have the same label,
        //but the second card of KK677 is stronger (K vs T), so KTJJT gets rank 2 and KK677 gets rank 3.
        assert_case(
            DealtHand::new("KK677", 20),
            DealtHand::new("KTJJT", 30),
            Ordering::Greater,
        );

        //So, 33332 and 2AAAA are both four of a kind hands,
        //but 33332 is stronger because its first card is stronger.
        assert_case(
            DealtHand::new("33332", 20),
            DealtHand::new("2AAAA", 30),
            Ordering::Greater,
        );

        //Similarly, 77888 and 77788 are both a full house,
        //but 77888 is stronger because its third card is stronger
        //(and both hands have the same first and second card).
        assert_case(
            DealtHand::new("77788", 30),
            DealtHand::new("77888", 20),
            Ordering::Less,
        );
        assert_case(
            DealtHand::new("19231", 30),
            DealtHand::new("18231", 20),
            Ordering::Greater,
        );
    }

    #[test]
    fn sort_example_correctly() {
        //32T3K is the only one pair and the other hands are all a stronger type, so it gets rank 1.
        //
        //KK677 and KTJJT are both two pair.
        //Their first cards both have the same label,
        //but the second card of KK677 is stronger (K vs T), so KTJJT gets rank 2 and KK677 gets rank 3.
        //
        //T55J5 and QQQJA are both three of a kind.
        //QQQJA has a stronger first card, so it gets rank 5 and T55J5 gets rank 4.

        let mut input: Vec<_> = vec![
            DealtHand::new("32T3K", 765),
            DealtHand::new("T55J5", 684),
            DealtHand::new("KK677", 28),
            DealtHand::new("KTJJT", 220),
            DealtHand::new("QQQJA", 483),
        ]
        .into_iter()
        .map(CategorizedHand::from)
        .collect();

        input.sort();

        insta::assert_debug_snapshot!(input);
    }
}
